from typing import Any, Tuple

import jax
import jax.numpy as jnp
from jax import Array

from .space import Space

Shape = Tuple[int, ...]


class Discrete(Space):
    """A discrete space consiting of :math:`\{0, 1, ..., n-1\}`."""

    _n: int

    def __init__(self, n: int):
        assert n >= 0
        self._n = n

    def sample(self, key: Array, shape: Shape = ()) -> Array:
        """
        Sample random action uniformly from :math:`\{0, 1, ..., n-1\}`.

        Arguments:
            key: A JAX random key used for sampling.
            shape: The shape of the returned sample.
        Returns:
            The sampled value.
        """
        return jax.random.randint(key, shape, 0, self.n)

    def contains(self, x: Any) -> Array:
        """
        Check whether specific object is within space.

        Arguments:
            x: The object to be checked.
        Returns:
            Whether the object is contained in the space.
        """
        return jnp.logical_and(x >= 0, x < self.n)

    @property
    def n(self) -> int:
        return self._n

    def __repr__(self) -> str:
        return f"Discrete({self.n})"



from abc import abstractmethod
from typing import Generic, TypeVar

import jax
import jax.numpy as jnp

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.typing import Action, Key


class GxmEnvironmentState(EnvironmentState):
    pass


TGxmEnvironmentState = TypeVar("TGxmEnvironmentState", bound=GxmEnvironmentState)


class GxmEnvironment(Generic[TGxmEnvironmentState], Environment[TGxmEnvironmentState]):
    """Base class for Gxm environments."""

    gxm_id: str
    """ The Gxm environment ID. """

    def init(self, key: Key) -> tuple[TGxmEnvironmentState, Timestep]:
        env_state, timestep = self._reset(key)
        return env_state, timestep

    def reset(
        self, key: Key, env_state: TGxmEnvironmentState
    ) -> tuple[TGxmEnvironmentState, Timestep]:
        env_state, timestep = self._reset(key)
        return env_state, timestep

    @abstractmethod
    def _reset(self, key: Key) -> tuple[TGxmEnvironmentState, Timestep]:
        pass

    def step(
        self, key: Key, env_state: TGxmEnvironmentState, action: Action
    ) -> tuple[TGxmEnvironmentState, Timestep]:
        env_state_step, timestep_step = self._step(key, env_state, action)
        env_state_reset, timestep_reset = self._reset(key)
        env_state = jax.tree.map(
            lambda x_step, x_reset: jnp.where(timestep_step.done, x_reset, x_step),
            env_state_step,
            env_state_reset,
        )
        obs = jax.tree.map(
            lambda x_step, x_reset: jnp.where(timestep_step.done, x_reset, x_step),
            timestep_step.obs,
            timestep_reset.obs,
        )
        true_obs = jax.tree.map(
            lambda x_step, x_reset: jnp.where(timestep_step.truncated, x_reset, x_step),
            timestep_step.obs,
            timestep_reset.obs,
        )
        timestep = Timestep(
            obs=obs,
            true_obs=true_obs,
            reward=timestep_step.reward,
            terminated=timestep_step.terminated,
            truncated=timestep_step.truncated,
            info=timestep_step.info,
        )
        return env_state, timestep

    @abstractmethod
    def _step(
        self, key: Key, env_state: TGxmEnvironmentState, action: Action
    ) -> tuple[TGxmEnvironmentState, Timestep]:
        pass



from typing import Any

import gymnax

from gxm.core import Environment


class GymnaxWrapper(gymnax.environments.environment.Environment):
    """A wrapper class for Gymnax environments."""

    env: Environment

    def __init__(self, env: Environment):
        self.env = env

    def reset(self, key: Any) -> Any:
        return None, self.env.reset(key)

    def step(self, key: Any, state: Any, action: Any) -> Any:
        next_state, reward, done, info = self.env.step(state, action)
        return next_state, reward, done, info

    def gxm_to_gymnax_space(self, gxm_space: Any) -> Any:
        """Convert a GXM space to a Gymnax space."""
        if gxm_space.__class__.__name__ == "Discrete":
            return gymnax.spaces.Discrete(gxm_space.n)
        elif gxm_space.__class__.__name__ == "Box":
            return gymnax.spaces.Box(
                low=gxm_space.low,
                high=gxm_space.high,
                shape=gxm_space.shape,
                dtype=gxm_space.dtype,
            )
        else:
            raise NotImplementedError(
                f"Conversion for {gxm_space.__class__.__name__} is not implemented."
            )



import jax.numpy as jnp

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper


class ClipReward(Wrapper):
    """
    Wrapper that clips the reward to a specified range.
    """

    env: Environment

    def __init__(
        self, env: Environment, unwrap: bool = True, min: float = -1.0, max: float = 1.0
    ):
        """
        Args:
            env: The environment to wrap.
            min: Minimum reward value.
            max: Maximum reward value.
        """
        super().__init__(env, unwrap=unwrap)
        self.min = min
        self.max = max

    def clip(self, reward: Array) -> Array:
        return jnp.clip(reward, self.min, self.max)

    def init(self, key: Key) -> tuple[EnvironmentState, Timestep]:
        env_state, timestep = self.env.init(key)
        timestep.reward = self.clip(timestep.reward)
        timestep.info["true_reward"] = timestep.reward
        return env_state, timestep

    def reset(
        self, key: Key, env_state: EnvironmentState
    ) -> tuple[EnvironmentState, Timestep]:
        env_state, timestep = self.env.reset(key, env_state)
        timestep.reward = self.clip(timestep.reward)
        timestep.info["true_reward"] = timestep.reward
        return env_state, timestep

    def step(
        self,
        key: Key,
        env_state: EnvironmentState,
        action: PyTree,
    ) -> tuple[EnvironmentState, Timestep]:
        env_state, timestep = self.env.step(key, env_state, action)
        timestep.reward = self.clip(timestep.reward)
        timestep.info["true_reward"] = timestep.reward
        return env_state, timestep



import jax.numpy as jnp
from jax import Array

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.wrappers.wrapper import Wrapper


class IgnoreTruncation(Wrapper):
    """
    A wrapper that treats truncation as termination and removes the corresponding obsercation from the timestep.

    >>> import gxm
    >>> from gxm.wrappers import IgnoreTruncation
    >>> env = make("Gymnax/CartPole-v1")
    >>> env = IgnoreTruncation(env)

    """

    env: Environment

    def __init__(self, env: Environment, actions: Array):
        """
        Args:
            env: The environment to wrap.
        """
        self.env = env

    def init(self, key: Array) -> tuple[EnvironmentState, Timestep]:
        env_state, timestep = self.env.init(key)
        timestep.truncated = None  # type: ignore
        timestep.true_obs = None  # type: ignore

        return env_state, timestep

    def reset(
        self, key: Array, env_state: EnvironmentState
    ) -> tuple[EnvironmentState, Timestep]:
        env_state, timestep = self.env.reset(key, env_state)
        timestep.truncated = None  # type: ignore
        timestep.true_obs = None  # type: ignore
        return env_state, timestep

    def step(
        self,
        key: Array,
        env_state: EnvironmentState,
        action: Array,
    ) -> tuple[EnvironmentState, Timestep]:

        env_state, timestep = self.env.step(key, env_state, action)
        timestep.terminated = jnp.logical_or(timestep.terminated, timestep.truncated)
        timestep.truncated = None  # type: ignore
        timestep.true_obs = None  # type: ignore
        return env_state, timestep



import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


class FlattenObservation(Wrapper):
    """Wrapper that adds a rollout method to the environment."""

    def __init__(self, env: Environment, unwrap: bool = True):
        super().__init__(env, unwrap=unwrap)

    @classmethod
    def flatten(cls, obs: PyTree) -> Array:
        obs_leaves = jax.tree.leaves(obs)
        obs_flat = jnp.concatenate([jnp.ravel(leaf) for leaf in obs_leaves])
        return obs_flat

    def init(self, key: Key) -> tuple[WrapperState, Timestep]:
        env_state, timestep = self.env.init(key)
        timestep.obs = self.flatten(timestep.obs)
        return env_state, timestep

    def reset(self, key: Key, env_state: WrapperState) -> tuple[WrapperState, Timestep]:
        env_state, timestep = self.env.reset(key, env_state)
        timestep.obs = self.flatten(timestep.obs)
        return env_state, timestep

    def step(
        self,
        key: Key,
        env_state: WrapperState,
        action: PyTree,
    ) -> tuple[WrapperState, Timestep]:
        env_state, timestep = self.env.step(key, env_state, action)
        timestep.obs = self.flatten(timestep.obs)
        return env_state, timestep



from dataclasses import dataclass
from typing import Any, Generic, TypeVar

import jax

from gxm.core import Environment, EnvironmentState


@jax.tree_util.register_dataclass
@dataclass
class WrapperState(EnvironmentState):
    env_state: EnvironmentState


TWrapperState = TypeVar("TWrapperState", bound=WrapperState)


class Wrapper(Generic[TWrapperState], Environment[TWrapperState]):
    """Base class for environment wrappers in gxm."""

    env: Environment
    unwrap: bool = True

    def __init__(self, env: Environment, unwrap: bool = True):
        self.env = env
        if isinstance(env, Wrapper) and not unwrap:
            assert not env.unwrap
        self.unwrap = unwrap

    def has_wrapper(self, wrapper_type: type[Environment]) -> bool:
        if isinstance(self, wrapper_type):
            return True
        return self.env.has_wrapper(wrapper_type)

    def get_wrapper(self, wrapper_type: type[Environment]) -> Environment:
        if isinstance(self, wrapper_type):
            return self
        return self.env.get_wrapper(wrapper_type)

    @property
    def unwrapped(self) -> Environment:
        if self.unwrap:
            return self.env.unwrapped
        return self

    def __getattr__(self, name: str) -> Any:
        if hasattr(self.env, name):
            return getattr(self.env, name)
        return getattr(self.env, name)



try:
    from gxm.environments.craftax_environment import CraftaxEnvironment
except ImportError:
    CraftaxEnvironment = None
try:
    from gxm.environments.gymnax_environment import GymnaxEnvironment
except ImportError:
    GymnaxEnvironment = None
try:
    from gxm.environments.pgx_environment import PgxEnvironment
except ImportError:
    PgxEnvironment = None
try:
    from gxm.environments.gymnasium_environment import GymnasiumEnvironment
except ImportError:
    GymnasiumEnvironment = None
try:
    from .envpool_environment import EnvpoolEnvironment
except:
    EnvpoolEnvironment = None
try:
    from gxm.environments.jaxatari_environmnet import JAXAtariEnvironment
except ImportError:
    JAXAtariEnvironment = None
try:
    from gxm.environments.xminigrid_environment import XMiniGridEnvironment
except ImportError:
    XMiniGridEnvironment = None
try:
    from gxm.environments.navix_environmnent import NavixEnvironment
except ImportError:
    NavixEnvironment = None

__all__ = [
    "GymnaxEnvironment",
    "PgxEnvironment",
    "CraftaxEnvironment",
    "EnvpoolEnvironment",
    "JAXAtariEnvironment",
    "GymnasiumEnvironment",
    "XMiniGridEnvironment",
    "NavixEnvironment",
]



from dataclasses import dataclass
from typing import Any

import gymnasium
import jax
import jax.numpy as jnp
import numpy as np

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Box, Discrete, Space, Tree
from gxm.typing import Array, Key

envs_gymnasium = {}


@jax.tree_util.register_dataclass
@dataclass
class GymnasiumState(EnvironmentState):
    env_id: Array


class GymnasiumEnvironment(Environment[GymnasiumState]):

    gymnasium_id: str
    """ The Gymnasium environment ID. """
    return_shape_dtype: Any
    """ The shape and dtype of the returned EnvironmentState and Timestep. """
    kwargs: Any
    """ The kwargs used to create the Gymnasium environment. """

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.gymnasium_id = id.split("/", 1)[1]
        env = gymnasium.make_vec(self.gymnasium_id, num_envs=1, **kwargs)
        obs, _ = env.reset()
        action = env.action_space.sample()
        obs, reward, terminated, truncated, info = env.step(action)
        env_state = GymnasiumState(env_id=jnp.int32(0))
        timestep = Timestep(
            obs=jnp.array(obs),
            true_obs=jnp.array(obs),
            reward=jnp.array(reward, dtype=jnp.float32),
            terminated=jnp.array(terminated, dtype=jnp.bool),
            truncated=jnp.array(truncated, dtype=jnp.bool),
            info=jax.tree.map(jnp.array, info),
        )
        self.return_shape_dtype = jax.tree.map(
            lambda x: jax.ShapeDtypeStruct(x.shape[1:], x.dtype), (env_state, timestep)
        )
        self.action_space = self.gymnasium_to_gxm_space(env.single_action_space)
        self.observation_space = self.gymnasium_to_gxm_space(
            env.single_observation_space
        )
        self.kwargs = kwargs

    def init(self, key: Key) -> tuple[GymnasiumState, Timestep]:
        def callback(key):
            global envs_gymnasium, current_env_id
            shape = key.shape[:-1]
            keys_flat = jnp.reshape(key, (-1, key.shape[-1]))
            num_envs = keys_flat.shape[0]
            envs = gymnasium.make_vec(
                self.gymnasium_id, num_envs=num_envs, **self.kwargs
            )
            obs, info = envs.reset(seed=0)
            env_id = len(envs_gymnasium)
            envs_gymnasium[env_id] = envs
            env_state = GymnasiumState(env_id=jnp.full(shape, env_id, dtype=jnp.int32))
            timestep = Timestep(
                obs=jnp.reshape(obs, shape + obs.shape[1:]),
                true_obs=jnp.reshape(obs, shape + obs.shape[1:]),
                reward=jnp.zeros(shape, dtype=jnp.float32),
                terminated=jnp.zeros(shape, dtype=jnp.bool),
                truncated=jnp.zeros(shape, dtype=jnp.bool),
                info=jax.tree.map(lambda i: jnp.reshape(i, shape + i.shape[1:]), info),
            )
            return env_state, timestep

        env_state, timestep = jax.pure_callback(
            callback,
            self.return_shape_dtype,
            jax.random.key_data(key),
            vmap_method="broadcast_all",
        )
        return env_state, timestep

    def reset(
        self, key: Key, env_state: GymnasiumState
    ) -> tuple[GymnasiumState, Timestep]:
        del key

        def callback(env_id):
            global envs_gymnasium
            shape = env_id.shape
            envs = envs_gymnasium[np.ravel(env_id)[0]]
            obs, info = envs.reset()
            env_state = GymnasiumState(env_id=jnp.full(shape, env_id, dtype=jnp.int32))
            timestep = Timestep(
                obs=jnp.reshape(obs, shape + obs.shape[1:]),
                true_obs=jnp.reshape(obs, shape + obs.shape[1:]),
                reward=jnp.zeros(shape, dtype=jnp.float32),
                terminated=jnp.zeros(shape, dtype=jnp.bool),
                truncated=jnp.zeros(shape, dtype=jnp.bool),
                info=jax.tree.map(lambda i: jnp.reshape(i, shape + i.shape[1:]), info),
            )
            return env_state, timestep

        env_state, timestep = jax.pure_callback(
            callback,
            self.return_shape_dtype,
            env_state.env_id,
            vmap_method="broadcast_all",
        )
        return env_state, timestep

    def step(
        self, key: Key, env_state: GymnasiumState, action: Array
    ) -> tuple[GymnasiumState, Timestep]:
        del key

        def callback(env_id, action):
            global envs_gymnasium
            shape = env_id.shape
            envs = envs_gymnasium[np.ravel(env_id)[0]]
            actions = np.reshape(np.asarray(action), (-1,))
            obs, reward, terminated, truncated, info = envs.step(actions)
            env_state = GymnasiumState(env_id=env_id)
            timestep = Timestep(
                obs=jnp.reshape(obs, shape + obs.shape[1:]),
                true_obs=jnp.reshape(obs, shape + obs.shape[1:]),
                reward=jnp.reshape(reward, shape).astype(jnp.float32),
                terminated=jnp.reshape(terminated, shape).astype(jnp.bool),
                truncated=jnp.reshape(truncated, shape).astype(jnp.bool),
                info=jax.tree.map(lambda i: jnp.reshape(i, shape + i.shape[1:]), info),
            )
            return env_state, timestep

        env_state, timestep = jax.pure_callback(
            callback,
            self.return_shape_dtype,
            env_state.env_id,
            action,
            vmap_method="broadcast_all",
        )

        return env_state, timestep

    @classmethod
    def gymnasium_to_gxm_space(cls, gymnasium_space: Any) -> Space:
        """Convert a Gymnasium space to a Gxm space."""
        if isinstance(gymnasium_space, gymnasium.spaces.Discrete):
            return Discrete(int(gymnasium_space.n))
        elif isinstance(gymnasium_space, gymnasium.spaces.Box):
            return Box(
                jnp.asarray(gymnasium_space.low),
                jnp.asarray(gymnasium_space.high),
                gymnasium_space.shape,
            )
        elif isinstance(gymnasium_space, gymnasium.spaces.MultiDiscrete):
            return Tree(tuple(Discrete(int(n)) for n in gymnasium_space.nvec))
        elif isinstance(gymnasium_space, gymnasium.spaces.Dict):
            return Tree(
                {
                    k: cls.gymnasium_to_gxm_space(v)
                    for k, v in gymnasium_space.spaces.items()
                }
            )
        elif isinstance(gymnasium_space, gymnasium.spaces.Tuple):
            return Tree(
                tuple(cls.gymnasium_to_gxm_space(s) for s in gymnasium_space.spaces)
            )
        else:
            raise NotImplementedError(
                f"Gymnasium space {gymnasium_space} not supported."
            )



from dataclasses import dataclass
from typing import cast

import jax
import pgx
from pgx.experimental import auto_reset

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Discrete
from gxm.typing import Array, Key


@jax.tree_util.register_dataclass
@dataclass
class PgxEnvironmentState(EnvironmentState):
    """State for Pgx environments."""

    pgx_state: pgx.State
    """The Pgx state."""


class PgxEnvironment(Environment[PgxEnvironmentState]):
    """Base class for Pgx environments."""

    pgx_id: str
    """The Pgx environment ID."""
    env: pgx.Env
    """The Pgx environment instance."""

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.pgx_id = id.split("/", 1)[1]
        self.env = pgx.make(cast(pgx.EnvId, self.pgx_id), **kwargs)
        self.action_space = Discrete(self.env.num_actions)

    def init(self, key: Key) -> tuple[PgxEnvironmentState, Timestep]:
        pgx_state = self.env.init(key)
        env_state = PgxEnvironmentState(pgx_state=pgx_state)
        timestep = Timestep(
            obs=pgx_state.observation,
            true_obs=pgx_state.observation,
            reward=pgx_state.rewards[pgx_state.current_player],
            terminated=pgx_state.terminated,
            truncated=pgx_state.truncated,
            info={},
        )
        return env_state, timestep

    def reset(
        self, key: Key, env_state: PgxEnvironmentState
    ) -> tuple[PgxEnvironmentState, Timestep]:
        del env_state
        return self.init(key)

    def step(
        self, key: Key, env_state: PgxEnvironmentState, action: Array
    ) -> tuple[PgxEnvironmentState, Timestep]:
        pgx_state = auto_reset(self.env.step, self.env.init)(
            env_state.pgx_state, action, key
        )
        env_state = PgxEnvironmentState(pgx_state=pgx_state)
        timestep = Timestep(
            obs=pgx_state.observation,
            true_obs=pgx_state.observation,
            reward=pgx_state.rewards[pgx_state.current_player],
            terminated=pgx_state.terminated,
            truncated=pgx_state.truncated,
            info={},
        )
        return env_state, timestep

    @property
    def num_actions(self) -> int:
        return self.env.num_actions



from dataclasses import dataclass
from typing import Sequence

import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class StackObservationsState(WrapperState):
    """State for the StackObservations wrapper."""

    obss: PyTree
    """The history of observations."""
    true_obss: PyTree
    """The history of true observations."""


class StackObservations(Wrapper[StackObservationsState]):
    """Wrapper that stacks the observation along a new axis."""

    num_stack: int
    padding: str

    def __init__(self, env: Environment, n_stack: int, padding: str = "reset"):
        self.env = env
        self.num_stack = n_stack
        self.padding = padding

    def init(self, key: Key) -> tuple[StackObservationsState, Timestep]:
        def stack(obss):
            return jax.tree.map(lambda *os: jnp.stack(os, axis=0), *obss)

        env_state, timestep = self.env.init(key)

        if self.padding == "reset":
            obss = stack(self.num_stack * [timestep.obs])
            true_obss = stack(self.num_stack * [timestep.true_obs])
            timestep.obs = obss
            timestep.true_obs = true_obss
        else:
            raise ValueError(f"Unknown padding method: {self.padding}")

        stack_observations_state = StackObservationsState(
            env_state=env_state, obss=obss, true_obss=true_obss
        )

        return stack_observations_state, timestep

    def reset(
        self, key: Key, env_state: StackObservationsState
    ) -> tuple[StackObservationsState, Timestep]:
        def stack(obss):
            return jax.tree.map(lambda *os: jnp.stack(os, axis=0), *obss)

        env_state, timestep = self.env.reset(key, env_state.env_state)

        if self.padding == "reset":
            obss = stack(self.num_stack * [timestep.obs])
            true_obss = stack(self.num_stack * [timestep.true_obs])
            timestep.obs = obss
            timestep.true_obs = true_obss
        else:
            raise ValueError(f"Unknown padding method: {self.padding}")

        stack_observations_state = StackObservationsState(
            env_state=env_state, obss=obss, true_obss=true_obss
        )

        return stack_observations_state, timestep

    def step(
        self,
        key: Key,
        env_state: StackObservationsState,
        action: PyTree,
    ) -> tuple[StackObservationsState, Timestep]:
        def concatenate(obss: Sequence[PyTree]) -> PyTree:
            return jax.tree.map(lambda *os: jnp.concatenate(os, axis=0), *obss)

        def expand_dims(obs: PyTree) -> PyTree:
            return jax.tree.map(lambda o: jnp.expand_dims(o, axis=0), obs)

        obss = env_state.obss
        true_obss = env_state.true_obss
        env_state, timestep = self.env.step(key, env_state.env_state, action)

        obss = jax.tree.map(lambda os: os[1:], obss)
        obss = concatenate([obss, expand_dims(timestep.obs)])
        true_obss = jax.tree.map(lambda tos: tos[1:], true_obss)
        true_obss = concatenate([true_obss, expand_dims(timestep.true_obs)])

        timestep.obs = obss
        timestep.true_obs = true_obss

        stack_observations_state = StackObservationsState(
            env_state=env_state,
            obss=obss,
            true_obss=true_obss,
        )

        return stack_observations_state, timestep



from dataclasses import dataclass

import jax
import jax.numpy as jnp
import navix

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Box, Discrete, Space
from gxm.typing import Action, Key


@jax.tree_util.register_dataclass
@dataclass
class NavixEnvironmentState(EnvironmentState):
    """State for Navix environments."""

    navix_state: navix.Timestep
    """The Navix state."""


class NavixEnvironment(Environment):
    """Base class for Gymnax environments."""

    navix_id: str
    """The Navix environment ID."""
    env: navix.Environment
    """The Navix environment instance."""

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.navix_id = id.split("/", 1)[1]
        self.env = navix.make(self.navix_id, **kwargs)
        self.action_space = self.navix_to_gxm_space(self.env.action_space)
        self.observation_space = self.navix_to_gxm_space(self.env.observation_space)

    def init(self, key: Key) -> tuple[NavixEnvironmentState, Timestep]:
        navix_state = self.env.reset(key)
        env_state = NavixEnvironmentState(navix_state=navix_state)
        timestep = Timestep(
            obs=navix_state.observation,
            true_obs=navix_state.observation,
            reward=jnp.float32(0.0),
            terminated=jnp.bool(True),
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep

    def reset(
        self, key: Key, env_state: NavixEnvironmentState
    ) -> tuple[NavixEnvironmentState, Timestep]:
        del env_state
        return self.init(key)

    def step(
        self, key: Key, env_state: NavixEnvironmentState, action: Action
    ) -> tuple[NavixEnvironmentState, Timestep]:
        del key
        navix_state = env_state.navix_state
        navix_state = self.env.step(navix_state, action)
        env_state = NavixEnvironmentState(navix_state=navix_state)
        timestep = Timestep(
            obs=navix_state.observation,
            true_obs=navix_state.observation,
            reward=navix_state.reward,
            terminated=navix_state.is_done(),
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep

    @property
    def num_actions(self) -> int:
        return len(self.env.action_set)

    @classmethod
    def navix_to_gxm_space(cls, navix_space) -> Space:
        """Convert a Navix space to a Gxm space."""
        if isinstance(navix_space, navix.spaces.Discrete):
            return Discrete(int(navix_space.n))
        if isinstance(navix_space, navix.spaces.Continuous):
            return Box(
                low=navix_space.minimum,
                high=navix_space.maximum,
                shape=navix_space.shape,
            )
        else:
            raise NotImplementedError(
                f"Gymnax space type {type(navix_space)} not supported."
            )



from dataclasses import dataclass

import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class StepCounterState(WrapperState):
    n_steps: Array


class StepCounter(Wrapper):
    """A wrapper that counts the number of steps taken in the environment."""

    def __init__(self, env: Environment, unwrap: bool = True):
        super().__init__(env, unwrap=unwrap)

    def init(self, key: Key) -> tuple[StepCounterState, Timestep]:
        env_state, timestep = self.env.init(key)
        step_counter_state = StepCounterState(
            env_state=env_state,
            n_steps=jnp.int32(0),
        )
        timestep.info["n_steps"] = step_counter_state.n_steps
        return step_counter_state, timestep

    def reset(
        self, key: Key, env_state: StepCounterState
    ) -> tuple[StepCounterState, Timestep]:
        step_counter_state = env_state
        env_state, timestep = self.env.reset(key, step_counter_state.env_state)
        step_counter_state = StepCounterState(
            env_state=env_state,
            n_steps=step_counter_state.n_steps,
        )
        timestep.info["n_steps"] = step_counter_state.n_steps
        return step_counter_state, timestep

    def step(
        self,
        key: Key,
        env_state: StepCounterState,
        action: PyTree,
    ) -> tuple[StepCounterState, Timestep]:
        step_counter_state = env_state
        env_state, timestep = self.env.step(key, env_state.env_state, action)
        step_counter_state = StepCounterState(
            env_state=env_state,
            n_steps=step_counter_state.n_steps + 1,
        )
        timestep.info["n_steps"] = step_counter_state.n_steps
        return step_counter_state, timestep



from dataclasses import dataclass

import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class StickyActionState(WrapperState):
    prev_action: PyTree


class StickyAction(Wrapper):
    """A wrapper that makes actions sticky with a given probability."""

    def __init__(self, env: Environment, unwrap: bool = True, stickiness: float = 0.25):
        super().__init__(env, unwrap=unwrap)
        self.stickiness = stickiness

    def init(self, key: Key) -> tuple[StickyActionState, Timestep]:
        env_state, timestep = self.env.init(key)
        sticky_action_state = StickyActionState(
            env_state=env_state,
            prev_action=self.env.action_space.sample(key),
        )
        return sticky_action_state, timestep

    def reset(
        self, key: Key, env_state: StickyActionState
    ) -> tuple[StickyActionState, Timestep]:
        env_state, timestep = self.env.reset(key, env_state)
        sticky_action_state = StickyActionState(
            env_state=env_state,
            prev_action=self.env.action_space.sample(key),
        )
        return sticky_action_state, timestep

    def step(
        self,
        key: Key,
        env_state: StickyActionState,
        action: PyTree,
    ) -> tuple[StickyActionState, Timestep]:
        sticky_action = jnp.where(
            jax.random.uniform(key) < self.stickiness,
            env_state.prev_action,
            action,
        )
        env_state, timestep = self.env.step(key, env_state.env_state, sticky_action)
        sticky_action_state = StickyActionState(
            env_state=env_state,
            prev_action=action,
        )
        return sticky_action_state, timestep



from typing import Any, Tuple

import jax
import jax.numpy as jnp
from jax import Array

from .space import Space

Shape = Tuple[int, ...]


class Box(Space):
    r"""A bounded box in :math:`\mathbb{R}^n`."""

    low: Array
    """Lower bound of the box."""
    high: Array
    """Upper bound of the box."""
    shape: Shape
    """Shape of the box."""

    def __init__(
        self,
        low: Array | float,
        high: Array | float,
        shape: Shape,
    ):
        self.low = jnp.broadcast_to(jnp.asarray(low, dtype=jnp.float32), shape)
        self.high = jnp.broadcast_to(jnp.asarray(high, dtype=jnp.float32), shape)
        self.shape = shape

    def sample(self, key: Array, shape: Shape = ()) -> Any:
        """
        Sample uniformly from the box.

        Args:
            key: JAX random key.
            shape: Shape of the sample to be drawn.
        Returns:
            Sample drawn from the box.
        """
        return jax.random.uniform(
            key, shape + self.shape, minval=self.low, maxval=self.high
        )

    def contains(self, x: Any) -> Array:
        """
        Check whether specific object is within space.

        Args:
            x: Object to be checked.
        Returns:
            Whether the object is within the space.
        """
        range_cond = jnp.logical_and(jnp.all(x >= self.low), jnp.all(x <= self.high))
        return range_cond

    @property
    def n(self) -> int:
        raise NotImplementedError("Box space does not have a single dimension 'n'.")

    def __repr__(self) -> str:
        return f"Box(low={self.low}, high={self.high}, shape={self.shape})"



from gxm.environments import (
    CraftaxEnvironment,
    EnvpoolEnvironment,
    GymnasiumEnvironment,
    GymnaxEnvironment,
    JAXAtariEnvironment,
    NavixEnvironment,
    PgxEnvironment,
    XMiniGridEnvironment,
)


def make(id: str, **kwargs):
    """
    Create an environment given its id.
    The id should be in the format "Library/EnvironmentName", e.g. "Gymnax/CartPole-v1".

    Args:
        id (str): The id of the environment to create.
        **kwargs: Additional keyword arguments to pass to the environment constructor.
    Returns:
        An instance of the requested environment.
    Raises:
        ValueError: If the library is not recognized.

    Examples:
        >>> env = make("Gymnax/CartPole-v1")
        >>> env = make("Pgx/MountainCarContinuous-v0")
        >>> env = make("Envpool/Pong-v5")
    """
    library = id.split("/", 1)[0]
    Environment = {
        "Gymnax": GymnaxEnvironment,
        "Pgx": PgxEnvironment,
        "Envpool": EnvpoolEnvironment,
        "Craftax": CraftaxEnvironment,
        "XMiniGrid": XMiniGridEnvironment,
        "JAXAtari": JAXAtariEnvironment,
        "Gymnasium": GymnasiumEnvironment,
        "Navix": NavixEnvironment,
    }[library]
    return Environment(id, **kwargs)



from gxm.core import Environment, EnvironmentState, Timestep, Trajectory, Transition
from gxm.registration import make

__all__ = [
    "Environment",
    "EnvironmentState",
    "Timestep",
    "Transition",
    "Trajectory",
    "make",
]



import envpool
import jax
import numpy as np

import gxm

num_envs = 8
id = "Breakout-v5"

env_gxm = gxm.make("Envpool/" + id)

key = jax.random.key(0)
keys = jax.random.split(key, num_envs)

env_envpool = envpool.make(id, env_type="gym", num_envs=num_envs)

env_states, timesteps = jax.vmap(env_gxm.init)(keys)
obs, _ = env_envpool.reset()

for _ in range(10000):
    actions = np.random.randint(0, 2, num_envs)
    env_states, timesteps = jax.vmap(env_gxm.step)(keys, env_states, actions)
    obs, rewards, dones, truncations, infos = env_envpool.step(actions)

    print(dones, timesteps.done)
    assert np.allclose(rewards, timesteps.reward)
    assert np.allclose(dones, timesteps.done)
    assert np.allclose(truncations, timesteps.truncated)
    assert np.allclose(obs, timesteps.obs)



"""Wrappers for ``gxm`` environments."""

from gxm.wrappers.clip_reward import ClipReward
from gxm.wrappers.discretize import Discretize
from gxm.wrappers.episode_counter import EpisodeCounter
from gxm.wrappers.episodic_life import EpisodicLife
from gxm.wrappers.evaluate import Evaluate
from gxm.wrappers.flatten_observation import FlattenObservation
from gxm.wrappers.ignore_truncation import IgnoreTruncation
from gxm.wrappers.record_episode_statistics import RecordEpisodeStatistics
from gxm.wrappers.rollout import Rollout
from gxm.wrappers.stack_observations import StackObservations
from gxm.wrappers.step_counter import StepCounter
from gxm.wrappers.sticky_action import StickyAction
from gxm.wrappers.wrapper import Wrapper

__all__ = [
    "ClipReward",
    "Discretize",
    "EpisodeCounter",
    "EpisodicLife",
    "Evaluate",
    "FlattenObservation",
    "IgnoreTruncation",
    "RecordEpisodeStatistics",
    "Rollout",
    "StackObservations",
    "StepCounter",
    "StickyAction",
    "Wrapper",
]



"""Spaces for defining action and observation spaces."""

from gxm.spaces.box import Box
from gxm.spaces.discrete import Discrete
from gxm.spaces.space import Space
from gxm.spaces.tree import Tree

__all__ = [
    "Discrete",
    "Box",
    "Tree",
    "Space",
]



import jax
import jax.numpy as jnp

from gxm.core import Environment, Trajectory
from gxm.typing import Key, Policy, PolicyState


def rollout(
    key: Key,
    env: Environment,
    pi: Policy,
    pi_state: PolicyState,
    n_steps: int,
) -> Trajectory:
    """
    Perform a rollout in the environment using the given policy.

    Args:
        key: A JAX random key for any stochasticity in the environment and policy.
        env: The environment to perform the rollout in.
        pi: The policy to use for selecting actions.
        pi_state: The initial state of the policy.
        n_steps: The number of steps to perform in the rollout.
    Returns:
        A Trajectory object containing the observations, actions, rewards, and other data from the rollout.
    """

    def step(carry, _):
        key, env_state, pi_state, timestep = carry
        key, key_pi, key_step = jax.random.split(key, 3)
        action, pi_state = pi(key_pi, pi_state, timestep.observation)
        env_state, timestep = env.step(key_step, env_state, action)
        carry = (key, env_state, pi_state, timestep)
        return carry, (timestep, action)

    env_state, timestep = env.init(key)
    carry = (key, env_state, pi_state, timestep)
    carry, (timesteps, actions) = jax.lax.scan(step, carry, None, length=n_steps)
    traj = timesteps.trajectory(timestep.obs, actions)
    return traj


def evaluate(
    key: Key,
    env: Environment,
    pi: Policy,
    pi_state: PolicyState,
    n_steps: int,
):
    """
    Evaluate a policy in the environment by performing a rollout and computing the mean return.

    Args:
        key: A JAX random key for any stochasticity in the environment and policy.
        env: The environment to perform the evaluation in.
        pi: The policy to evaluate.
        pi_state: The initial state of the policy.
        n_steps: The number of steps to perform in the rollout.
    Returns:
        The mean return of the policy over the rollout.
    """
    traj = rollout(key, env, pi, pi_state, n_steps=n_steps)

    def step(carry, x):
        r = carry
        reward, done = x
        r = reward + (1.0 - done) * r
        return r, r

    rs, _ = jax.lax.scan(step, 0.0, (traj.reward, traj.done))
    mean_return = jnp.sum(traj.done * rs) / jnp.sum(traj.done)
    return mean_return



from dataclasses import dataclass

import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class EpisodeStatistics:
    episode_length: Array
    episodic_return: Array
    episodic_discounted_return: Array
    mask: Array


@jax.tree_util.register_dataclass
@dataclass
class CurrentStatistics:
    current_return: Array
    current_length: Array
    current_discounted_return: Array


@jax.tree_util.register_dataclass
@dataclass
class RecordEpisodeStatisticsState(WrapperState):
    current_stats: CurrentStatistics
    episode_stats: EpisodeStatistics

    def __iter__(self):
        yield self.env_state
        yield self.current_stats
        yield self.episode_stats


class RecordEpisodeStatistics(Wrapper[RecordEpisodeStatisticsState]):
    """
    A wrapper that records the episode length :math:`T` , episodic return
    :math:`J(\\tau) = \\sum_{t=0}^{T} r_t` , and discounted episodic return
    :math:`G(\\tau) = \\sum_{t=0}^{T} \\gamma^t r_t` at the end of each episode.
    The statistics can be accessed from the ``info`` field of the ``Timestep`` returned
    by the environment. It will contain the stats of the most recent finished episode.
    By default , the discount factor :math:`\\gamma` is set to 1.0, meaning that the
    episodic return and discounted episodic return are the same.
    """

    gamma: float
    """The discount factor :math:`\\gamma` for calculating the discounted episodic return."""
    n_episodes: int
    """The number of past episodes to record statistics for."""

    def __init__(
        self,
        env: Environment,
        unwrap: bool = True,
        gamma: float = 1.0,
        n_episodes: int = 1,
    ):
        super().__init__(env, unwrap=unwrap)
        self.gamma = gamma
        self.n_episodes = n_episodes

    @staticmethod
    def get_averaged_stats(episode_stats: EpisodeStatistics) -> dict[str, jax.Array]:
        valid_episodes = jnp.maximum(jnp.sum(episode_stats.mask), 1.0)
        episode_length = (
            jnp.sum(episode_stats.episode_length * episode_stats.mask) / valid_episodes
        )
        episodic_return = (
            jnp.sum(episode_stats.episodic_return * episode_stats.mask) / valid_episodes
        )
        episodic_discounted_return = (
            jnp.sum(episode_stats.episodic_discounted_return * episode_stats.mask)
            / valid_episodes
        )
        return {
            "episode_length": episode_length,
            "episodic_return": episodic_return,
            "episodic_discounted_return": episodic_discounted_return,
        }

    def init(self, key: Key) -> tuple[RecordEpisodeStatisticsState, Timestep]:
        env_state, timestep = self.env.init(key)
        current_stats = CurrentStatistics(
            current_return=jnp.float32(0.0),
            current_length=jnp.int32(0.0),
            current_discounted_return=jnp.float32(0.0),
        )
        episode_stats = EpisodeStatistics(
            episodic_return=jnp.zeros(self.n_episodes, dtype=jnp.float32),
            episodic_discounted_return=jnp.zeros(self.n_episodes, dtype=jnp.float32),
            episode_length=jnp.zeros(self.n_episodes, dtype=jnp.int32),
            mask=jnp.zeros(self.n_episodes, dtype=jnp.int32),
        )
        record_episode_stats_state = RecordEpisodeStatisticsState(
            env_state=env_state,
            current_stats=current_stats,
            episode_stats=episode_stats,
        )
        timestep.info |= {
            "current_length": current_stats.current_length,
            "current_return": current_stats.current_return,
            "current_discounted_return": current_stats.current_discounted_return,
        }
        timestep.info |= self.get_averaged_stats(episode_stats)
        return record_episode_stats_state, timestep

    def reset(
        self, key: Key, env_state: RecordEpisodeStatisticsState
    ) -> tuple[RecordEpisodeStatisticsState, Timestep]:
        env_state, timestep = self.env.reset(key, env_state.env_state)
        current_stats = CurrentStatistics(
            current_return=jnp.float32(0.0),
            current_length=jnp.int32(0.0),
            current_discounted_return=jnp.float32(0.0),
        )
        episode_stats = EpisodeStatistics(
            episodic_return=jnp.zeros(self.n_episodes, dtype=jnp.float32),
            episodic_discounted_return=jnp.zeros(self.n_episodes, dtype=jnp.float32),
            episode_length=jnp.zeros(self.n_episodes, dtype=jnp.int32),
            mask=jnp.zeros(self.n_episodes, dtype=jnp.int32),
        )
        record_episode_stats_state = RecordEpisodeStatisticsState(
            env_state=env_state,
            current_stats=current_stats,
            episode_stats=episode_stats,
        )
        timestep.info |= {
            "current_length": current_stats.current_length,
            "episode_length": episode_stats.episode_length,
            "current_return": current_stats.current_return,
        }
        timestep.info |= self.get_averaged_stats(episode_stats)
        return record_episode_stats_state, timestep

    def step(
        self,
        key: Key,
        env_state: RecordEpisodeStatisticsState,
        action: PyTree,
    ) -> tuple[RecordEpisodeStatisticsState, Timestep]:
        wrapper_state = env_state
        current_stats = wrapper_state.current_stats
        episode_stats = wrapper_state.episode_stats

        env_state, timestep = self.env.step(key, wrapper_state.env_state, action)

        done = timestep.done
        reward = timestep.reward

        current_return = current_stats.current_return + reward
        current_discounted_return = (
            current_stats.current_discounted_return
            + reward * self.gamma**current_stats.current_length
        )
        current_length = current_stats.current_length + 1

        episodic_return = jnp.where(
            done,
            jnp.roll(episode_stats.episodic_return, 1).at[0].set(current_return),
            episode_stats.episodic_return,
        )
        episodic_discounted_return = jnp.where(
            done,
            jnp.roll(episode_stats.episodic_discounted_return, 1)
            .at[0]
            .set(current_discounted_return),
            episode_stats.episodic_discounted_return,
        )
        episode_length = jnp.where(
            done,
            jnp.roll(episode_stats.episode_length, 1).at[0].set(current_length),
            episode_stats.episode_length,
        )
        mask = jnp.where(
            done,
            jnp.roll(episode_stats.mask, 1).at[0].set(1),
            episode_stats.mask,
        )

        current_return = (1 - done) * current_return
        current_discounted_return = (1 - done) * current_discounted_return
        current_length = (1 - done) * current_length

        current_stats = CurrentStatistics(
            current_return=current_return,
            current_length=current_length,
            current_discounted_return=current_discounted_return,
        )
        episode_stats = EpisodeStatistics(
            episode_length=episode_length,
            episodic_return=episodic_return,
            episodic_discounted_return=episodic_discounted_return,
            mask=mask,
        )
        wrapper_state = RecordEpisodeStatisticsState(
            env_state=env_state,
            current_stats=current_stats,
            episode_stats=episode_stats,
        )
        timestep.info |= {
            "current_length": current_stats.current_length,
            "current_return": current_stats.current_return,
            "current_discounted_return": current_stats.current_discounted_return,
        }
        timestep.info |= self.get_averaged_stats(episode_stats)

        return wrapper_state, timestep



from typing import Any, Callable, TypeAlias

import jax

Array: TypeAlias = jax.Array
Key: TypeAlias = jax.Array
PyTree: TypeAlias = Any

Action: TypeAlias = PyTree
Observation: TypeAlias = PyTree

PolicyState: TypeAlias = PyTree
Policy: TypeAlias = Callable[[Key, PolicyState, Observation], Action]

QState: TypeAlias = PyTree
Q: TypeAlias = Callable[[Key, QState, Observation, Action], Array]



from dataclasses import dataclass

import jax
import jax.numpy as jnp
import xminigrid
import xminigrid.environment

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Discrete, Tree
from gxm.typing import Key


@jax.tree_util.register_dataclass
@dataclass
class XMiniGridEnvironmentState(EnvironmentState):
    """State for XMiniGrid environments."""

    xminigrid_state: xminigrid.environment.TimeStep
    """ The XMiniGrid environment state. """


class XMiniGridEnvironment(Environment[XMiniGridEnvironmentState]):
    """Base class for XMiniGrid environments."""

    xminigrid_id: str
    """The XMiniGrid environment ID."""
    env: xminigrid.environment.Environment
    """The XMiniGrid environment instance."""
    env_params: xminigrid.environment.EnvParams
    """The parameters for the XMiniGrid environment."""

    def __init__(self, id: str, **kwargs):
        self.id = "XMiniGrid/" + id
        self.xminigrid_id = id.split("/", 1)[1]
        self.env, self.env_params = xminigrid.make(self.xminigrid_id, **kwargs)
        self.action_space = Discrete(self.env.num_actions(self.env_params))
        observation_shape = self.env.observation_shape(self.env_params)
        assert type(observation_shape) is tuple
        self.observation_space = Tree(tuple(Discrete(n) for n in observation_shape))

    def init(self, key: Key) -> tuple[XMiniGridEnvironmentState, Timestep]:
        xminigrid_state = self.env.reset(self.env_params, key)
        env_state = XMiniGridEnvironmentState(xminigrid_state=xminigrid_state)
        timestep = Timestep(
            obs=xminigrid_state.observation,
            true_obs=xminigrid_state.observation,
            reward=jnp.float32(0.0),
            terminated=jnp.bool(False),
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep

    def reset(
        self, key: jax.Array, env_state: XMiniGridEnvironmentState
    ) -> tuple[XMiniGridEnvironmentState, Timestep]:
        del env_state
        return self.init(key)

    def step(
        self, key: jax.Array, env_state: XMiniGridEnvironmentState, action: jax.Array
    ) -> tuple[XMiniGridEnvironmentState, Timestep]:
        del key
        xminigrid_state = env_state.xminigrid_state
        xminigrid_state = self.env.step(self.env_params, xminigrid_state, action)
        env_state = XMiniGridEnvironmentState(xminigrid_state=xminigrid_state)
        timestep = Timestep(
            obs=xminigrid_state.observation,
            true_obs=xminigrid_state.observation,
            reward=xminigrid_state.reward,
            terminated=xminigrid_state.last(),
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep



from dataclasses import dataclass
from typing import Any

import jax
import jax.numpy as jnp
import jaxatari
import jaxatari.core
import jaxatari.spaces
import jaxatari.wrappers
from jaxatari.wrappers import AtariWrapper, PixelObsWrapper

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Box, Discrete, Space, Tree
from gxm.typing import Array, Key


@jax.tree_util.register_dataclass
@dataclass
class JAXAtariEnvironmentState(EnvironmentState):
    """State for JAXAtari environments."""

    jaxatari_state: Any
    """State for JAXAtari environments."""


class JAXAtariEnvironment(Environment[JAXAtariEnvironmentState]):
    """Base class for JAXAtari environments."""

    jaxatari_id: str
    """The JAXAtari environment id."""
    env: jaxatari.wrappers.JaxatariWrapper
    """The JAXAtari environment."""
    env_params: Any
    """The JAXAtari environment parameters."""

    def __init__(self, id: str, **kwargs):

        self.id = id
        self.jaxatari_id = id.split("/", 1)[1]
        env = jaxatari.core.make(self.jaxatari_id, **kwargs)
        env = AtariWrapper(env)
        self.env = PixelObsWrapper(env)
        self.action_space = self.jaxatari_to_gxm_space(self.env.action_space())

    def init(self, key: Key) -> tuple[JAXAtariEnvironmentState, Timestep]:
        obs, jaxatari_state = self.env.reset(key)
        obs = self.to_grayscale(obs)
        obs = self.resize(obs)
        env_state = JAXAtariEnvironmentState(jaxatari_state=jaxatari_state)
        timestep = Timestep(
            obs=obs,
            true_obs=obs,
            reward=jnp.float32(0.0),
            terminated=jnp.bool(False),
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep

    def reset(
        self, key: Key, env_state: JAXAtariEnvironmentState
    ) -> tuple[JAXAtariEnvironmentState, Timestep]:
        del env_state
        return self.init(key)

    def step(
        self, key: Key, env_state: JAXAtariEnvironmentState, action: Array
    ) -> tuple[JAXAtariEnvironmentState, Timestep]:
        del key
        jaxatari_state = env_state.jaxatari_state
        obs, jaxatari_state, reward, done, _ = self.env.step(jaxatari_state, action)
        obs = self.to_grayscale(obs)
        obs = self.resize(obs)
        env_state = JAXAtariEnvironmentState(jaxatari_state=jaxatari_state)
        timestep = Timestep(
            obs=obs,
            true_obs=obs,
            reward=jnp.float32(reward),
            terminated=jnp.bool(done),
            truncated=jnp.bool(done),
            info={},
        )
        return env_state, timestep

    @classmethod
    def to_grayscale(cls, obs: Array) -> Array:
        """Convert an RGB observation to grayscale."""
        return jnp.dot(obs[..., :3], jnp.array([0.2989, 0.5870, 0.1140]))

    @classmethod
    def resize(cls, obs: Array) -> Array:
        """Resize an observation to 84x84."""
        return jax.image.resize(obs, (obs.shape[0], 84, 84), method="bilinear")

    @classmethod
    def jaxatari_to_gxm_space(cls, jaxatari_space) -> Space:
        """Convert a Gymnax space to a Gxm space."""
        if isinstance(jaxatari_space, jaxatari.spaces.Discrete):
            return Discrete(jaxatari_space.n)
        if isinstance(jaxatari_space, jaxatari.spaces.Box):
            return Box(
                low=jaxatari_space.low,
                high=jaxatari_space.high,
                shape=jaxatari_space.shape,
            )
        if isinstance(jaxatari_space, jaxatari.spaces.Dict):
            return Tree(
                {
                    k: cls.jaxatari_to_gxm_space(v)
                    for k, v in jaxatari_space.spaces.items()
                }
            )
        else:
            raise NotImplementedError(
                f"JAXAtari space type {type(jaxatari_space)} not supported."
            )



from dataclasses import dataclass

import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class EvaluateState(WrapperState):
    current_return: Array
    cumulative_return: Array
    n_episodes: Array

    @property
    def mean_return(self) -> Array:
        return self.cumulative_return / self.n_episodes


class Evaluate(Wrapper[EvaluateState]):
    env: Environment

    def __init__(self, env: Environment, unwrap: bool = True):
        """
        Args:
            env: The environment to wrap.
            unwrap: Whether to unwrap the environment or treat it as part of the base environment.
        """
        super().__init__(env, unwrap=unwrap)

    def init(self, key: Key) -> tuple[EvaluateState, Timestep]:
        env_state, timestep = self.env.init(key)
        evaluate_state = EvaluateState(
            env_state=env_state,
            current_return=jax.numpy.zeros(timestep.reward.shape),
            cumulative_return=jax.numpy.zeros(timestep.reward.shape),
            n_episodes=jax.numpy.zeros(timestep.reward.shape),
        )
        return evaluate_state, timestep

    def reset(
        self, key: Key, env_state: EvaluateState
    ) -> tuple[EvaluateState, Timestep]:
        evaluate_state = env_state
        env_state, timestep = self.env.reset(key, evaluate_state.env_state)
        evaluate_state = EvaluateState(
            env_state=env_state,
            current_return=jax.numpy.zeros(timestep.reward.shape),
            cumulative_return=jax.numpy.zeros(timestep.reward.shape),
            n_episodes=jax.numpy.zeros(timestep.reward.shape),
        )
        return evaluate_state, timestep

    def step(
        self,
        key: Key,
        env_state: EvaluateState,
        action: PyTree,
    ) -> tuple[EvaluateState, Timestep]:
        evaluate_state = env_state
        env_state, timestep = self.env.step(key, evaluate_state.env_state, action)
        current_return = evaluate_state.current_return + timestep.reward
        cumulative_return = jnp.where(
            timestep.done,
            evaluate_state.cumulative_return + current_return,
            evaluate_state.cumulative_return,
        )
        n_episodes = jnp.where(
            timestep.done,
            evaluate_state.n_episodes + 1,
            evaluate_state.n_episodes,
        )
        current_return = jnp.where(
            timestep.done,
            jnp.zeros_like(current_return),
            current_return,
        )
        evaluate_state = EvaluateState(
            env_state=env_state,
            current_return=current_return,
            cumulative_return=cumulative_return,
            n_episodes=n_episodes,
        )
        return evaluate_state, timestep



from dataclasses import dataclass
from typing import Any

import gymnax
import gymnax.environments.spaces
import jax
import jax.numpy as jnp

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Box, Discrete, Space, Tree
from gxm.typing import Array, Key


@jax.tree_util.register_dataclass
@dataclass
class GymnaxEnvironmentState(EnvironmentState):
    """State for Gymnax environments."""

    gymnax_state: gymnax.EnvState
    """ The Gymnax environment state. """


class GymnaxEnvironment(Environment):
    """Base class for Gymnax environments."""

    gymnax_id: str
    """ The Gymnax environment ID. """
    env: gymnax.environments.environment.Environment
    """ The Gymnax environment. """
    env_params: Any
    """ The parameters for the Gymnax environment. """

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.gymnax_id = id.split("/", 1)[1]
        self.env, self.env_params = gymnax.make(self.gymnax_id, **kwargs)
        self.action_space = self.gymnax_to_gxm_space(
            self.env.action_space(self.env_params)
        )
        self.observation_space = self.gymnax_to_gxm_space(
            self.env.observation_space(self.env_params)
        )

    def init(self, key: Key) -> tuple[GymnaxEnvironmentState, Timestep]:
        obs, gxm_state = self.env.reset(key, self.env_params)
        env_state = GymnaxEnvironmentState(gymnax_state=gxm_state)
        timestep = Timestep(
            obs=obs,
            true_obs=obs,
            reward=jnp.float32(0.0),
            terminated=jnp.bool(True),
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep

    def reset(
        self, key: Key, env_state: GymnaxEnvironmentState
    ) -> tuple[GymnaxEnvironmentState, Timestep]:
        del env_state
        return self.init(key)

    def step(
        self, key: Key, env_state: GymnaxEnvironmentState, action: Array
    ) -> tuple[GymnaxEnvironmentState, Timestep]:
        gymnax_state = env_state.gymnax_state
        obs, gymnax_state, reward, done, _ = self.env.step(
            key, gymnax_state, action, self.env_params
        )
        env_state = GymnaxEnvironmentState(gymnax_state=gymnax_state)
        timestep = Timestep(
            obs=obs,
            true_obs=obs,
            reward=reward,
            terminated=done,
            truncated=jnp.bool(False),
            info={},
        )
        return env_state, timestep

    @classmethod
    def gymnax_to_gxm_space(cls, gymnax_space) -> Space:
        """Convert a Gymnax space to a Gxm space."""
        if isinstance(gymnax_space, gymnax.environments.spaces.Discrete):
            return Discrete(gymnax_space.n)
        if isinstance(gymnax_space, gymnax.environments.spaces.Box):
            return Box(
                low=gymnax_space.low,
                high=gymnax_space.high,
                shape=gymnax_space.shape,
            )
        if isinstance(gymnax_space, gymnax.environments.spaces.Dict):
            return Tree(
                {k: cls.gymnax_to_gxm_space(v) for k, v in gymnax_space.spaces.items()}
            )
        if isinstance(gymnax_space, gymnax.environments.spaces.Tuple):
            return Tree([cls.gymnax_to_gxm_space(s) for s in gymnax_space.spaces])
        else:
            raise NotImplementedError(
                f"Gymnax space type {type(gymnax_space)} not supported."
            )



from dataclasses import dataclass

import jax
from jax import numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class EpisodicLifeState(WrapperState):
    lives: Array


class EpisodicLife(Wrapper[EpisodicLifeState]):
    """
    A wrapper that makes losing a life in an environment (like Atari games) count as the end of an episode.
    It assumes that the environment's timestep info dictionary contains a "lives" key indicating the number of lives remaining.
    """

    env: Environment

    def __init__(self, env: Environment):
        """
        Args:
            env: The environment to wrap.
        """
        self.env = env

    def init(self, key: Key) -> tuple[EpisodicLifeState, Timestep]:
        env_state, timestep = self.env.init(key)
        lives = timestep.info["lives"]
        episodic_life_state = EpisodicLifeState(env_state=env_state, lives=lives)
        return episodic_life_state, timestep

    def reset(
        self, key: Key, env_state: EpisodicLifeState
    ) -> tuple[EpisodicLifeState, Timestep]:
        env_state, timestep = self.env.reset(key, env_state.env_state)
        lives = timestep.info["lives"]
        episodic_life_state = EpisodicLifeState(env_state=env_state, lives=lives)
        return episodic_life_state, timestep

    def step(
        self,
        key: Key,
        env_state: EpisodicLifeState,
        action: PyTree,
    ) -> tuple[EpisodicLifeState, Timestep]:
        prev_lives = env_state.lives
        env_state, timestep = self.env.step(key, env_state.env_state, action)
        lives = timestep.info["lives"]
        episodic_life_state = EpisodicLifeState(env_state=env_state, lives=lives)
        timestep.terminated = jnp.logical_or(timestep.terminated, lives < prev_lives)
        return episodic_life_state, timestep



from dataclasses import dataclass

import jax
import jax.numpy as jnp

from gxm.core import Environment, Timestep
from gxm.typing import Array, Key, PyTree
from gxm.wrappers.wrapper import Wrapper, WrapperState


@jax.tree_util.register_dataclass
@dataclass
class EpisodeCounterState(WrapperState):
    n_episodes: Array


class EpisodeCounter(Wrapper):
    """A wrapper that counts the number of episodes completed in the environment."""

    def __init__(self, env: Environment, unwrap: bool = True):
        super().__init__(env, unwrap=unwrap)

    def init(self, key: Key) -> tuple[EpisodeCounterState, Timestep]:
        env_state, timestep = self.env.init(key)
        episode_counter_state = EpisodeCounterState(
            env_state=env_state,
            n_episodes=jnp.int32(0),
        )
        timestep.info["n_episodes"] = episode_counter_state.n_episodes
        return episode_counter_state, timestep

    def reset(
        self, key: Key, env_state: EpisodeCounterState
    ) -> tuple[EpisodeCounterState, Timestep]:
        episode_counter_state = env_state
        env_state, timestep = self.env.reset(key, episode_counter_state.env_state)
        episode_counter_state = EpisodeCounterState(
            env_state=env_state,
            n_episodes=episode_counter_state.n_episodes,
        )
        timestep.info["n_episodes"] = episode_counter_state.n_episodes
        return episode_counter_state, timestep

    def step(
        self,
        key: Key,
        env_state: EpisodeCounterState,
        action: PyTree,
    ) -> tuple[EpisodeCounterState, Timestep]:
        episode_counter_state = env_state
        env_state, timestep = self.env.step(key, env_state.env_state, action)
        episode_counter_state = EpisodeCounterState(
            env_state=env_state,
            n_episodes=jnp.where(
                timestep.done,
                episode_counter_state.n_episodes + 1,
                episode_counter_state.n_episodes,
            ),
        )
        timestep.info["n_episodes"] = episode_counter_state.n_episodes
        return episode_counter_state, timestep



from typing import Any, Tuple

from jax import Array

Shape = Tuple[int, ...]


class Space:
    """Abstract base class for action and observation spaces."""

    def sample(self, key: Array, shape: Shape = ()) -> Any:
        del key, shape
        raise NotImplementedError

    def contains(self, x: Array) -> Any:
        del x
        raise NotImplementedError

    @property
    def n(self) -> int:
        raise NotImplementedError



from dataclasses import dataclass
from typing import Any

import gymnax.environments.spaces
import jax
import jax.numpy as jnp
from craftax.craftax_env import make_craftax_env_from_name

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.spaces import Box, Discrete, Space, Tree
from gxm.typing import Action, Key


@jax.tree_util.register_dataclass
@dataclass
class CraftaxEnvironmentState(EnvironmentState):
    """State for Craftax environments."""

    craftax_state: Any
    """ The Craftax environment state. """


class CraftaxEnvironment(Environment[CraftaxEnvironmentState]):
    """Base class for Craftax environments."""

    craftax_id: str
    """ The Craftax environment ID. """
    env: Any
    """ The Craftax environment. """
    env_params = Any
    """ The parameters for the Craftax environment. """

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.craftax_id = id.split("/", 1)[1]
        self.env = make_craftax_env_from_name(
            self.craftax_id, auto_reset=True, **kwargs
        )
        self.env_params = self.env.default_params
        self.action_space = self.craftax_to_gxm_space(
            self.env.action_space(self.env_params)
        )
        self.observation_space = self.craftax_to_gxm_space(
            self.env.observation_space(self.env_params)
        )

    def init(self, key: Key) -> tuple[CraftaxEnvironmentState, Timestep]:
        obs, craftax_state = self.env.reset(key, self.env_params)
        obs, _, _, _, info = self.env.step(
            key, craftax_state, jnp.array(0), self.env_params
        )
        env_state = CraftaxEnvironmentState(craftax_state=craftax_state)
        timestep = Timestep(
            obs=obs,
            true_obs=obs,
            reward=jnp.float32(0.0),
            terminated=jnp.bool(True),
            truncated=jnp.bool(False),
            info=info,
        )
        return env_state, timestep

    def reset(
        self, key: Key, env_state: CraftaxEnvironmentState
    ) -> tuple[CraftaxEnvironmentState, Timestep]:
        del env_state
        return self.init(key)

    def step(
        self, key: Key, env_state: CraftaxEnvironmentState, action: Action
    ) -> tuple[CraftaxEnvironmentState, Timestep]:
        craftax_state = env_state.craftax_state
        obs, craftax_state, reward, done, info = self.env.step(
            key, craftax_state, action, self.env_params
        )
        env_state = CraftaxEnvironmentState(craftax_state=craftax_state)
        timestep = Timestep(
            obs=obs,
            true_obs=obs,
            reward=reward,
            terminated=done,
            truncated=done,
            info=info,
        )
        return env_state, timestep

    @classmethod
    def craftax_to_gxm_space(cls, gymnax_space) -> Space:
        """Convert a Craftax space to a Gxm space."""
        if isinstance(gymnax_space, gymnax.environments.spaces.Discrete):
            return Discrete(gymnax_space.n)
        if isinstance(gymnax_space, gymnax.environments.spaces.Box):
            return Box(
                low=gymnax_space.low,
                high=gymnax_space.high,
                shape=gymnax_space.shape,
            )
        if isinstance(gymnax_space, gymnax.environments.spaces.Dict):
            return Tree(
                {k: cls.craftax_to_gxm_space(v) for k, v in gymnax_space.spaces.items()}
            )
        if isinstance(gymnax_space, gymnax.environments.spaces.Tuple):
            return Tree([cls.craftax_to_gxm_space(s) for s in gymnax_space.spaces])
        else:
            raise NotImplementedError(
                f"Craftax space type {type(gymnax_space)} not supported."
            )



from typing import Any, Tuple

import jax
import jax.numpy as jnp

from .space import Space

Shape = Tuple[int, ...]


class Tree(Space):
    """Space composed of multiple subspaces in a tree structure."""

    def __init__(self, spaces: Any):
        self.spaces = spaces

    def sample(self, key: jax.Array, shape: Shape = ()) -> Any:
        """
        Sample random action from all subspaces, retaining the original structure.

        Args:
            key: JAX random key.
            shape: Shape of the sample to be drawn from each subspace.
        Returns:
            A tree-structured object with samples from each subspace.
        """
        structure = jax.tree.structure(
            self.spaces, is_leaf=lambda x: isinstance(x, Space)
        )
        keys = jax.random.split(key, structure.num_leaves)
        key_tree = jax.tree.unflatten(structure, keys)
        return jax.tree.map(
            lambda space, k: space.sample(k, shape), self.spaces, key_tree
        )

    def contains(self, x: jax.Array) -> bool:
        """
        Check whether dimensions of object are within subspace.

        Args:
            x: Object to be checked.
        Returns:
            True if object is within all subspaces, False otherwise.
        """
        structure = jax.tree.structure(
            self.spaces, is_leaf=lambda x: isinstance(x, Space)
        )
        x_structure = jax.tree.structure(
            x, is_leaf=lambda y: isinstance(y, (jnp.ndarray, int, float))
        )
        if structure != x_structure:
            return False
        return jax.tree.reduce(
            lambda a, b: a and b,
            jax.tree.map(lambda space, y: space.contains(y), self.spaces, x),
            True,
        )

    @property
    def n(self) -> int:
        return jax.tree.reduce(
            lambda a, b: a * b.n,
            self.spaces,
            1,
            lambda x: isinstance(x, Space),
        )

    def __repr__(self) -> str:
        return f"Tree({self.spaces})"



import jax

from gxm.core import Environment, EnvironmentState, Timestep
from gxm.typing import Key, PyTree
from gxm.wrappers.wrapper import Wrapper


class Discretize(Wrapper):
    """
    Wrapper that discretizes a continuous action space.
    Maps a discrete set of actions to the continuous action space of the environment.
    The actions are specified as a list of continuous actions :math:`A`.
    The action space of the wrapped environment is then :math:`\{0, 1, \ldots, |A|-1\}`.

    >>> import gxm
    >>> from gxm.wrappers import Discretize
    >>> env = make("Gymnasium/Pendulum-v1")
    >>> actions = jnp.array([-2.0, 0.0, 2.0])
    >>> env = Discretize(env, actions)

    The actions passed to the ``Discretize`` wrapper need to be of shape :math:`(|A|, D)`,
    where :math:`|A|` is the number of discrete actions and :math:`D` is the dimensionality of the
    continuous action space of the wrapped environment.
    """

    env: Environment
    actions: PyTree

    def __init__(self, env: Environment, actions: PyTree, unwrap: bool = True):
        """
        Args:
            env: The environment to wrap.
            actions: The discrete set of actions to map to.
            unwrap: Whether to unwrap the environment or treat it as part of the base environment.
        """
        super().__init__(env, unwrap=unwrap)
        self.actions = actions

    def init(self, key: Key) -> tuple[EnvironmentState, Timestep]:
        return self.env.init(key)

    def reset(
        self, key: Key, env_state: EnvironmentState
    ) -> tuple[EnvironmentState, Timestep]:
        return self.env.reset(key, env_state)

    def step(
        self,
        key: Key,
        env_state: EnvironmentState,
        action: PyTree,
    ) -> tuple[EnvironmentState, Timestep]:
        action = jax.tree.map(lambda x: x[action], self.actions)
        return self.env.step(key, env_state, action)



from typing import Callable

import jax
from jax import Array

from gxm.core import Environment, EnvironmentState, Timestep, Trajectory
from gxm.wrappers.wrapper import Wrapper


class Rollout(Wrapper):
    """Wrapper that adds a rollout method to the environment."""

    def __init__(self, env: Environment):
        self.env = env

    def init(self, key: Array) -> tuple[EnvironmentState, Timestep]:
        return self.env.init(key)

    def reset(
        self, key: Array, env_state: EnvironmentState
    ) -> tuple[EnvironmentState, Timestep]:
        return self.env.reset(key, env_state)

    def step(
        self,
        key: Array,
        env_state: EnvironmentState,
        action: Array,
    ) -> tuple[EnvironmentState, Timestep]:
        return self.env.step(key, env_state, action)

    def rollout(
        self,
        key: Array,
        env_state: EnvironmentState,
        pi: Callable[[Array, Array], Array],
        num_steps: int,
    ) -> tuple[EnvironmentState, Trajectory]:
        def _step(carry, key):
            env_state, prev_obs = carry
            key_pi, key_step = jax.random.split(key)
            action = pi(
                key_pi,
                prev_obs,
            )
            env_state, timestep = self.step(key_step, env_state, action)
            return (env_state, timestep.obs), (env_state.timestep, action)

        first_obs = env_state.obs
        keys = jax.random.split(key, num_steps)
        env_state, (timesteps, actions) = jax.lax.scan(
            _step,
            env_state,
            keys,
        )
        trajectory = timesteps.trajectory(first_obs, actions)
        return env_state, trajectory



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

import jax
import jax.numpy as jnp

from gxm.spaces import Space
from gxm.typing import Array, Key, PyTree


@jax.tree_util.register_dataclass
@dataclass
class Timestep:
    """
    Class representing a single timestep :math:`(R_i, S_{i+1})` in an environment.
    Where :math:`R_i` is the reward received after taking an action at timestep
    :math:`i` and :math:`S_{i+1}` is the observation at the next timestep.
    In case of truncation, ``true_obs`` represents the observation :math:`\hat{S}_{i+1}` that would have been
    observed if the episode had not been truncated.
    """

    reward: Array
    """The reward :math:`R_i` received at this timestep :math:`i`."""
    terminated: Array
    """Whether the episode has terminated at this timestep."""
    truncated: Array
    """Whether the episode has been truncated at this timestep."""
    obs: PyTree
    """The observation :math:`S_{i+1}` at this timestep :math:`i`."""
    true_obs: PyTree
    """The true observation :math:`\hat{S}_{i+1}` at this timestep. This may differ from ``obs`` in environments that allow truncation, if and only if truncation is True. """
    info: dict[str, PyTree]
    """Additional information about the timestep."""

    @property
    def done(self) -> Array:
        """Whether the episode has terminated or been truncated."""
        return jnp.logical_or(self.terminated, self.truncated)

    def transition(
        self,
        prev_obs: PyTree,
        action: Array,
        prev_info: dict[str, Any] = {},
    ) -> "Transition":
        """Convert the current timestep :math:`(R_t, S_{t+1})` into a transition
        :math:`(S_t, A_t, R_t, S_{t+1})` given the previous observation :math:`S_t`
        and the action :math:`A_t`.
        Args:
            prev_obs: The observation at the previous timestep.
            action: The action taken at the current timestep.
            prev_info: The info at the previous timestep.
        Returns:
            A Transition object containing the current and next timesteps.
        """
        return Transition(
            prev_obs=prev_obs,
            prev_info=prev_info,
            action=action,
            reward=self.reward,
            terminated=self.terminated,
            truncated=self.truncated,
            obs=self.obs,
            info=self.info,
        )

    def trajectory(
        self, first_obs: PyTree, action: Array, first_info: dict[str, PyTree] = {}
    ) -> "Trajectory":
        r"""
        Convert a sequence of timesteps :math:`(R_0, S_1, ..., S_n)` with
        the first observation :math:`S_0` and the actions :math:`(A_0, A_1, ..., A_{n-1})`
        into a trajectory :math:`(S_0, A_0, R_0, S_1, ..., S_n)`.

        Args:
            first_obs: The observation at the first timestep.
            action: The action taken at each timestep.
        Returns:
            A Trajectory object containing the sequence of timesteps.
        """
        assert (
            self.obs.shape[0] == action.shape[0]
        ), "The number of observations must match the number of actions."
        return Trajectory(
            obs=jnp.concatenate([first_obs[None], self.obs], axis=0),
            true_obs=jnp.concatenate([first_obs[None], self.true_obs], axis=0),
            reward=self.reward,
            terminated=self.terminated,
            truncated=self.truncated,
            info=self.info,
            action=action,
        )


@jax.tree_util.register_dataclass
@dataclass
class Transition:
    """Class representing a single transition :math:`(S_i, A_i, R_i, S_{i+1})` in an environment."""

    prev_obs: PyTree
    prev_info: dict[str, PyTree]
    action: PyTree
    reward: Array
    terminated: Array
    truncated: Array
    obs: PyTree
    info: dict[str, PyTree]

    @property
    def done(self) -> Array:
        """Return whether the episode has ended (either terminated or truncated)."""
        return jnp.logical_or(self.terminated, self.truncated)


@jax.tree_util.register_dataclass
@dataclass
class Trajectory:
    """Class representing a trajectory :math:`(S_0, A_0, R_0, S_1, ..., S_n)` in an environment."""

    obs: PyTree
    """The observations :math:`(S_0, S_1, ..., S_n)` in the trajectory."""
    true_obs: PyTree
    """The true observations :math:`(\hat{S}_0, \hat{S}_1, ..., \hat{S}_n)` in the trajectory. These may differ from ``obs`` in environments that allow truncation."""
    action: PyTree
    """The actions :math:`(A_0, A_1, ..., A_{n-1})` taken in the trajectory."""
    reward: Array
    """The rewards :math:`(R_0, R_1, ..., R_{n-1})` received in the trajectory."""
    terminated: Array
    """Whether the episode terminated at each timestep in the trajectory."""
    truncated: Array
    """Whether the episode was truncated at each timestep in the trajectory."""
    info: dict[str, PyTree]
    """Additional information about the trajectory."""

    @property
    def done(self) -> Array:
        """Return whether the episode has ended (either terminated or truncated)."""
        return jnp.logical_or(self.terminated, self.truncated)

    def __len__(self):
        """Return the length of the trajectory."""
        assert (
            self.reward.ndim == 1
        ), "Trajectory length is only defined for batch size 1."
        return self.reward.shape[0]


class EnvironmentState:
    """
    A placeholder class for environment state.
    This can be replaced with a more specific implementation as needed.
    """

    pass


TEnvironmentState = TypeVar("TEnvironmentState", bound=EnvironmentState)


class Environment(Generic[TEnvironmentState], ABC):
    """
    Base class for environments in ``gxm``.
    Environments should inherit from this class and implement the
    ``init``, ``step``, ``reset``, and ``num_actions`` methods.
    """

    id: str
    """The unique identifier of the environment."""
    action_space: Space
    """The action space of the environment."""
    observation_space: Space
    """The observation space of the environment."""

    @abstractmethod
    def init(self, key: Key) -> tuple[TEnvironmentState, Timestep]:
        """
        Initialize the environment and return the initial state.

        Args:
            key: A JAX random key for any stochastic initialization.
        Returns:
            A tuple containing the initial environment state and the initial timestep.
        """

    @abstractmethod
    def reset(
        self, key: Key, env_state: TEnvironmentState
    ) -> tuple[TEnvironmentState, Timestep]:
        """
        Reset the environment to its initial state.

        Args:
            key: A JAX random key for any stochasticity in the environment.
            env_state: The current state of the environment.
        Returns:
            A tuple containing the reset environment state and the initial timestep.
        """
        pass

    @abstractmethod
    def step(
        self,
        key: Key,
        env_state: TEnvironmentState,
        action: PyTree,
    ) -> tuple[TEnvironmentState, Timestep]:
        """
        Perform a step in the environment given an action.

        Args:
            key: A JAX random key for any stochasticity in the environment.
            env_state: The current state of the environment.
            action: The action to take in the environment.
        Returns:
            A tuple containing the new environment state and the resulting timestep.
        """
        pass

    def has_wrapper(self, wrapper_type: type["Environment"]) -> bool:
        """
        Check if the environment or any of its wrappers is of a specific type.

        Args:
            wrapper_type: The type to check for.
        Returns:
            True if the environment or any of its wrappers is of the specified type, False otherwise.
        """
        return isinstance(self, wrapper_type)

    def get_wrapper(self, wrapper_type: type["Environment"]) -> "Environment":
        """
        Retrieve the first wrapper of a specific type from the environment.

        Args:
            wrapper_type: The type of the wrapper to retrieve.
        Returns:
            The first wrapper of the specified type.
        Raises:
            ValueError: If no wrapper of the specified type is found.
        """
        if isinstance(self, wrapper_type):
            return self
        raise ValueError(f"No wrapper of type {wrapper_type} found in the environment.")

    @property
    def unwrapped(self) -> "Environment":
        """
        Retrieve the base environment by unwrapping all wrappers.

        Returns:
            The base environment without any wrappers.
        """
        return self
